package ips.audio.coreaudio;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.LineEvent;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.SourceDataLine;
import javax.sound.sampled.TargetDataLine;

public class CoreAudioSourceDataLine extends CoreAudioDataLine implements
		SourceDataLine {

	public static boolean DEFAULT_CLONE_MISSING_CHANNELS=true;
	public static boolean DEFAULT_AUTO_MONO_TO_STEREO=true;
	private boolean cloneMissingChannesl=DEFAULT_CLONE_MISSING_CHANNELS;
	
	
	private boolean autoMonoToStereo=DEFAULT_AUTO_MONO_TO_STEREO;
	
	private boolean stopped=false;
	public CoreAudioSourceDataLine(CoreAudioMixer mixer,ByteBuffer dataline) {
		super(mixer,dataline);
		AudioFormat[] suppFormats=getSupportedFormats().toArray(new AudioFormat[0]);
	     try{
	    	 lineInfo=new DataLine.Info(SourceDataLine.class,suppFormats,-1,-1);
	     }catch(IllegalArgumentException iae){
	         if(DEBUG)System.out.println("Illegal data line ");  
	     }
	}


	
	/**
	 * Returns the normalized float value (amplitude) of the given data.
	 * 
	 * @param sample audio sample data buffer
 	 * @param offset offset in the buffer
	 * @return normalized float value (-1.0 ... +1.0)
	 */
	public float getFloatValue(byte[] sample, int offset) {

		if (signed) {
			int is;
			if (!bigEndian) {

				is = (int) sample[offset + sampleSize - 1];
				for (int i = sampleSize - 2; i >= 0; i--) {
					is = is << 8;
					is = is | (0xFF & (int) sample[offset + i]);
				}
			} else {
				is = (int) sample[offset];
				for (int i = 0; i < sampleSize-1; i++) {
					is = is << 8;
					is = is | (0xFF & (int) sample[i + offset]);
				}
			}
			return (float) is / (float)maxValue;
		} else {
			// Not tested !!!
			long is;
			if (!bigEndian) {

				is = 0xFF & (long) sample[offset + sampleSize - 1];
				for (int i = sampleSize - 2; i >= 0; i--) {
					is = is << 8;
					is = is | (0xFF & (long) sample[offset + i]);
				}
			} else {
				is = 0xFF & (long) sample[offset];
				for (int i = 0; i < sampleSize-1; i++) {
					is = is << 8;
					is = is | (0xFF & (long) sample[i + offset]);
				}
			}
			return (float) is / (float)maxValue;
		}
	}
	
	
	private native void addToWritePos(ByteBuffer dataline,int written);
	@Override
	public int write(byte[] b, int off, int len) {
		
				
				
				
				int toWrite=len;
				int offset=off;
				while(toWrite>0 && !stopped){
					int avail=available();
					int toWrite1=toWrite;
					if(toWrite1 > avail){
						toWrite1=avail;
					}
					long writePos=getWritePos(dataline);
					long bufWritePos=writePos%coreAudioBufferSize;
					long rbToWrite1=jsByteCountToCoreAudioByteCount(toWrite1);
					if(bufWritePos+rbToWrite1 > coreAudioBufferSize){
						rbToWrite1=coreAudioBufferSize-bufWritePos;
					}
					long floatStartIndex=bufWritePos/COREAUDIO_FLOAT_SAMPLESIZE;
					long floatLen=rbToWrite1/COREAUDIO_FLOAT_SAMPLESIZE;
					
					float v=0.0f;
					for(long i=0;i<floatLen;i++){
						int currPCh=(int)i%physicalChannels;
						if(currPCh<channels){
							v =getFloatValue(b, offset);
							offset+=sampleSize;
						}else if(autoMonoToStereo && channels==1 && currPCh==1){
							// cloned (previous value)
						}else{
							v=0.0f;
						}
						ringBufferFloat.put((int) (floatStartIndex+i), v);
						bufWritePos+=COREAUDIO_FLOAT_SAMPLESIZE;
					}
					toWrite-=coreAudioByteCountToJsByteCount(rbToWrite1);
					addToWritePos(dataline,(int)rbToWrite1);
				}
				
				
				return offset-off;
	}
	public void start() {
		stopped=false;
		super.start();
	}
	@Override
	public void stop() {
		stopped=true;
		super.stop();
	}

	private native void drain(ByteBuffer dataline);
	private native boolean drained(ByteBuffer dataline);
	@Override
	public void drain() {
		drain(dataline);
		while(!stopped && !drained(dataline) ){
		try {
			Thread.sleep(100);
		} catch (InterruptedException e) {
			// OK
		}
		}
	}

	private native void flush(ByteBuffer dataline);
	@Override
	public void flush() {
		flush(dataline);
	}

	@Override
	public long getLongFramePosition() {
		if(!open)return 0;
		long caReadPos=getReadPos(dataline);
		return coreAudioByteCountToJsByteCount(caReadPos)/audioFormat.getFrameSize();
	}

	private native int coreAudioBytesAvailable(ByteBuffer jdl);
	@Override
	public int available() {
		int caAvailBytes=coreAudioBytesAvailable(dataline);
		return (int)coreAudioByteCountToJsByteCount(caAvailBytes);

	}
}
